__version__ = "0.5.4"
"""
:mod:`tobler` --- A library for spatial interpolation
=================================================

"""
from . import area_weighted
from . import dasymetric
from . import model
